
use serenity::{
    client::Context, 
    framework::standard::{
        Args, 
        Reason, 
        CommandOptions, 
        macros::check
    }, 
    model::channel::Message
};

use crate::{
    func::*
};

#[check]
#[name("DraftID")]
pub(crate) async fn id_check(ctx: &Context, msg: &Message, args: &mut Args, _: &CommandOptions) -> Result<(), Reason> {
    match ctx.filter_draft_id(args.current()).await {
        Ok(_) => {
            Ok(())
        },
        Err(y) => {
            match y {
                DraftParseError::Invalid => {
                    Err(Reason::User("That doesn't seem to be a valid draft ID. Check it for misspellings.".into()))
                },
                DraftParseError::Missing => {
                    match msg.guild_id {
                        Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                            Some(_) => Ok(()),
                            None => {
                                Err(Reason::User("You must specify a draft ID!".into()))
                            }
                        },
                        None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                            Some(_) => Ok(()),
                            None => {
                                Err(Reason::User("You must specify a draft ID!".into()))
                            }
                        }
                    }
                }
            }
        }
    }
}

#[check]
#[name("DraftID2")]
pub(crate) async fn id_check_permissive(ctx: &Context, msg: &Message, args: &mut Args, _: &CommandOptions) -> Result<(), Reason> {
    match ctx.filter_draft_id(args.current()).await {
        Ok(_) => {
            Ok(())
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(_) => Ok(()),
                    None => {
                        Err(Reason::User("You must specify a draft ID!".into()))
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(_) => Ok(()),
                    None => {
                        Err(Reason::User("You must specify a draft ID!".into()))
                    }
                }
            }
        }
    }
}

#[check]
#[name("Joined")]
pub(crate) async fn joined_check(ctx: &Context, msg: &Message, args: &mut Args, _: &CommandOptions) -> Result<(), Reason> {
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(s) => {
            s
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        return Err(Reason::User("You must specify a draft ID!".into()));
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        return Err(Reason::User("You must specify a draft ID!".into()));
                    }
                }
            }
        }
    };

    let mod_role = ctx.get_mod_role(draft_id.clone()).await;
    let allowed = {
        if mod_role.is_some() {
            let mod_role = mod_role.unwrap();
            let g = ctx.get_server_id(draft_id.clone()).await;
            ctx.is_owner(draft_id.clone(), msg.author.id.as_u64()).await || msg.author.has_role(ctx, g, mod_role).await.unwrap()
        }
        else {
            ctx.is_owner(draft_id.clone(), msg.author.id.as_u64()).await
        }
    };

    if allowed {
        return Ok(())
    }

    if !ctx.check_draft_joined(draft_id, msg.author.id.as_u64()).await {
        return Err(Reason::User("You're not signed up in that draft!".into()));
    }

    Ok(())
}

#[check]
#[name("Manager")]
pub(crate) async fn dman_check(ctx: &Context, msg: &Message, args: &mut Args, _: &CommandOptions) -> Result<(), Reason> {
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        return Err(Reason::User("You must specify a draft ID!".into()));
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        return Err(Reason::User("You must specify a draft ID!".into()));
                    }
                }
            }
        }
    };

    let mod_role = ctx.get_mod_role(draft_id.clone()).await;
    let allowed = {
        if mod_role.is_some() {
            let mod_role = mod_role.unwrap();
            let g = ctx.get_server_id(draft_id.clone()).await;
            ctx.is_owner(draft_id.clone(), msg.author.id.as_u64()).await || msg.author.has_role(ctx, g, mod_role).await.unwrap()
        }
        else {
            ctx.is_owner(draft_id.clone(), msg.author.id.as_u64()).await
        }
    };

    match allowed {
        true => Ok(()),
        false => {
            return Err(Reason::User("You are not a moderator of this draft!".into()));
        }
    }
}

#[check]
#[name("DMsOrAdmin")]
pub(crate) async fn dm_or_admin(ctx: &Context, msg: &Message, _: &mut Args, _: &CommandOptions) -> Result<(), Reason> {
    let admin = if let Some(g) = msg.guild(ctx) {
        let m = g.member(ctx, msg.author.id).await.expect("failed member get");
        m.permissions(ctx).expect("permissions please.").administrator()
    }
    else {
        false
    };

    match admin || msg.is_private() {
        true => Ok(()),
        false => {
            Err(Reason::User("You are not an administrator of this server!".into()))
        }
    }
}