use flaken::Flaken;
use log::*;
use pokerust::{Ability, FlavorText, FromName, Pokemon, PokemonForm, PokemonSpecies, PokemonSpeciesVariety};
use rand::prelude::SliceRandom;
use serenity::{client::Context, framework::standard::{Args, CommandResult, macros::{command}}, model::channel::Message};

use crate::{DoublesTierList, EMBED_COLOR, SmogonTierList, TYPE_NAMES, emoji_from_type_id, 
    data::PokemonFormExt,
    func::{
        GenesectContext, 
        page
    }
};

#[command]
#[description("Look up basic information about a Pokémon species.")]
#[aliases("dex")]
#[usage("<species name>")]
async fn pokedex(ctx: &Context, msg: &Message, args: Args) -> CommandResult {
    info!("Pokedex Initiated");
    let pot_pokemon = args.rest().to_lowercase();
    let pokemon = Pokemon::from_name(pot_pokemon.as_str());
    let form = PokemonForm::from_name_exhaustive(pot_pokemon.as_str());
    debug!("Attempting to retrieve Pokedex for {}", pot_pokemon);
    let (pokemon, species, form) = match pokemon {
        Ok(p) => {
            let s = match p.species.get() {
                Ok(s) => s,
                Err(why) => {
                    error!("Pokedex: species failed: {}", why);
                    return Ok(())
                }
            };
            let f = match form {
                Ok(f) => f,
                Err(_) => {
                    if let Some(f) = p.forms.first() {
                        match f.get() {
                            Ok(f) => f,
                            Err(x) => {
                                error!("Pokedex: form failed: {}", x);
                                return Ok(())
                            }
                        }
                    }
                    else {
                        error!("Pokedex: form failed. no forms associated with pokemon");
                        return Ok(())
                    }
                }
            };
            debug!("Pokedex: full retrieval success (mon #{}, spec #{}, form #{})", p.id, s.id, f.id);
            
            (p, s, f)
        },
        Err(y) => {
            match PokemonSpecies::from_name(pot_pokemon.as_str()) {
                Ok(s) => {
                    warn!("Pokedex: form failed: {}; falling back to species", y);
                    let v: Vec<&PokemonSpeciesVariety> = s.varieties.iter().filter(|p| p.is_default).collect();
                    let p = v.first().unwrap().pokemon.get()?;
                    let f = match form {
                        Ok(f) => f,
                        Err(_) => {
                            if let Some(f) = p.forms.first() {
                                match f.get() {
                                    Ok(f) => f,
                                    Err(x) => {
                                        error!("Pokedex: form failed: {}", x);
                                        return Ok(())
                                    }
                                }
                            }
                            else {
                                error!("Pokedex: form failed. no forms associated with pokemon");
                                return Ok(())
                            }
                        }
                    };
                    debug!("Pokedex: full retrieval success (mon #{}, spec #{}, form #{})", p.id, s.id, f.id);

                    (p, s, f)
                },
                Err(z) => {
                    match form {
                        Ok(f) => {
                            let p = f.pokemon.get()?;
                            let s = p.species.get()?;
                            debug!("Pokedex: full retrieval success (mon #{}, spec #{}, form #{})", p.id, s.id, f.id);

                            (p, s, f)
                        },
                        Err(_) => {
                            error!("Pokedex: all failed: {}; {}", y, z);
                            let _ = msg.reply(ctx, "That's probably not a Pokémon species!").await;
                            return Ok(())
                        }
                    }
                }
            }
        }
    };
    let dexno = &species.pokedex_numbers.iter().filter(|x| x.pokedex.name == "national").next().map_or(species.id as u16, |v| v.entry_number);
    debug!("Pokedex: #{}", dexno);
    let name = &species.names.iter().filter(|x| x.language.name == "en").next().map_or(&species.name, |v| &v.name);
    debug!("Pokedex: {}", name);
    let mut genera = species.genera.iter().filter(|x| x.language.name == "en");
    let genus = genera.next().map_or("??? Pokémon", |v| &v.genus);
    debug!("Pokedex: Genus {}", genus);
    let form_name = &form.form_names.iter().filter(|x| x.language.name == "en").next();
    let entries: Vec<&FlavorText> = species.flavor_text_entries.iter().filter(|x| x.language.name == "en").collect();
    let entry = entries.choose(&mut rand::thread_rng());
    debug!("Pokedex: entry? {}", entry.is_some());
    let sprite = pokemon.sprites.other.official_artwork.front_default.map_or(
        format!("https://xayah.net/static/pokemon/{}.png", pokemon.id),
        |v| v
    );
    debug!("Pokedex: official art link: {}", sprite);
    /*let bhp = pokemon.stats.iter().filter(|x| x.stat.name == "hp").next().map_or(0, |v| v.base_stat);
    let batk = pokemon.stats.iter().filter(|x| x.stat.name == "attack").next().map_or(0, |v| v.base_stat);
    let bdef = pokemon.stats.iter().filter(|x| x.stat.name == "defense").next().map_or(0, |v| v.base_stat);
    let bspa = pokemon.stats.iter().filter(|x| x.stat.name == "special-attack").next().map_or(0, |v| v.base_stat);
    let bspd = pokemon.stats.iter().filter(|x| x.stat.name == "special-defense").next().map_or(0, |v| v.base_stat);
    let bspe = pokemon.stats.iter().filter(|x| x.stat.name == "speed").next().map_or(0, |v| v.base_stat);*/
    let tier = {
        let data = ctx.data.read().await;
        let tiers = data.get::<SmogonTierList>().expect("tiers missing");
        if tiers.contains_key(&form.id) {
            Some(tiers[&form.id].clone())
        }
        else {
            None
        }
    };
    debug!("got tier");
    let d_tier = {
        let data = ctx.data.read().await;
        let tiers = data.get::<DoublesTierList>().expect("tiers missing");
        if tiers.contains_key(&form.id) {
            Some(tiers[&form.id].clone())
        }
        else {
            None
        }
    };
    debug!("got doubles tier");
    let tier_title = if tier.is_some() {
        if d_tier.is_some() {
            Some(format!("{} / {}", tier.unwrap(), d_tier.unwrap()))
        } else { Some(format!("{} / Illegal", tier.unwrap())) }
    } else { None };
    let title = if form_name.is_some() {
        format!("#{:03}: {} ({})", dexno, name, form_name.unwrap().name)
    } else {
        format!("#{:03}: {}", dexno, name)
    };
    debug!("tiertitle");
    debug!("{:?}; {:?}", pokemon.types[0], pokemon.types[0].type_);
    debug!("{:?}", pokemon.types[0].type_.get());
    let t1 = pokemon.types[0].type_.get()?.id as u8;
    debug!("got type 1");
    let t2 = {
        if pokemon.types.len() > 1 {
            let i = pokemon.types[1].type_.get()?.id as u8;
            Some(i)
        }
        else { None }
    };
    debug!("got types");
    let type_string = {
        let mut s = String::new();
        let e1 = emoji_from_type_id(t1);
        s.push_str(e1);
        s.push_str(" ");
        s.push_str(TYPE_NAMES[t1 as usize]);
        if t2.is_some() {
            let t2 = t2.unwrap();
            let e2 = emoji_from_type_id(t2);
            s.push_str(" / ");
            s.push_str(e2);
            s.push_str(" ");
            s.push_str(TYPE_NAMES[t2 as usize]);
        }
        s
    };

    let abilities = {
        let m: Vec<(u8, Ability)> = pokemon.abilities.iter().map(|a| {
            (a.slot, a.ability.get())
        })
        .filter(|(_, a)| {
            a.is_ok()
        })
        .map(|(s, a)| {
            (s, a.unwrap())
        }).collect();
        let mut split: Vec<(u8, String, String)> = m.iter().map(|(s, a)| {
            let n = a.names.iter().filter(|a| a.language.name == "en").next();
            let f = a.flavor_text_entries.iter().filter(|a| a.language.name == "en").next();
            (
                s.to_owned(),
                if n.is_some() { n.unwrap().name.clone() } else { a.name.clone() },
                if f.is_some() { f.unwrap().flavor_text.clone() } else { "Flavor text missing".to_owned() }
            )
        }).collect();
        split.sort_by(|(a, _, _), (b, _, _)| a.cmp(b));
        split
    };
    debug!("got abilities");

    let statgon = form.stat_hexagon();
    debug!("got statgon");

    let msg = msg.channel_id.send_message(&ctx, |m| {
        m.embed(|e| {
            e.title(&title);
            e.description(genus);
            e.thumbnail(sprite);
            e.image(statgon);
            
            if tier_title.is_some() {
                e.field("Smogon Tiers (Singles / Doubles)", tier_title.unwrap(), false);
            }
            e.field("Type", type_string, false);
            /*e.fields(vec![
                ("Base HP", format!("{}", bhp), true),
                ("Base Attack", format!("{}", batk), true),
                ("Base Defense", format!("{}", bdef), true),
                ("Base Sp. Attack", format!("{}", bspa), true),
                ("Base Sp. Def.", format!("{}", bspd), true),
                ("Base Speed", format!("{}", bspe), true),
            ]);*/
            for (s, k, v) in abilities {
                match s {
                    1 => {
                        e.field(format!("Ability 1: {}", k), v, false);
                    },
                    2 => {
                        e.field(format!("Ability 2: {}", k), v, false);
                    },
                    3 => {
                        e.field(format!("Hidden Ability: {}", k), v, false);
                    },
                    _ => ()
                }
            }
            if entry.is_some() {
                let ntr = entry.unwrap();
                let version = {
                    let v = ntr.version.clone().unwrap();
                    let v = v.get().expect("failed");
                    let q = v.names.iter().filter(|a| a.language.name == "en").next();
                    match q {
                        Some(n) => n.name.clone(),
                        None => {
                            let n = v.name.clone();
                            let mut c: Vec<char> = n.chars().collect();
                            c[0].make_ascii_uppercase();
                            let n: String = c.iter().collect();
                            n
                        }
                    }
                };
                let text = ntr.flavor_text.replace('\n', " ");
                e.field(&format!("Pokédex Entry ({})", version), text, false);
            }  
            e.colour(EMBED_COLOR);
            e.url(&format!("https://bulbapedia.bulbagarden.net/wiki/{}_(Pokémon)", name.replace(" ", "_")));

            e
        });
        m
    }).await;
    debug!("sent message");

    if let Err(why) = msg {
        error!("Error sending message: {:?}", why);
    }

    Ok(())
}

#[command]
async fn id(ctx: &Context, msg: &Message) -> CommandResult {
    let mut flake = Flaken::default().node(1).epoch(1613940319944);
    let uuid = flake.next();
    let bytes = uuid.to_be_bytes();
    let strs: Vec<String> = bytes.iter()
                               .map(|b| format!("{:02X}", b))
                               .collect();
    let hex = strs.join(" ");
    let id = mnemonic::to_string(bytes);

    let _ = msg.channel_id.say(ctx, format!("{}\n{}\n{}", id, hex, uuid)).await;

    Ok(())
}

#[command]
#[description("List the current drafts.")]
#[usage("[all]")]
async fn drafts(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft List Initiated");
    let all = match args.single::<String>() {
        Ok(m) => {
            let v = m.to_lowercase();
            v == "all" || msg.is_private()
        },
        Err(_) => {
            msg.is_private()
        }
    };
    if all {
        match ctx.get_drafts().await {
            Ok(b) => {
                let _ = page(ctx, msg, b, "List of Drafts".to_owned(), None).await;
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }
    else {
        match ctx.get_drafts_in_server(msg.guild_id.unwrap().as_u64()).await {
            Ok(b) => {
                let s_name = msg.guild_id.unwrap().name(ctx).unwrap();
                let _ = page(ctx, msg, b, format!("List of Drafts in {}", s_name), None).await;
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }

    Ok(())
}
