use std::{
    collections::{
        HashMap, 
        HashSet
    }, 
    env, 
    fs::File, 
    fs::{
        self, 
        OpenOptions
    }, 
    io::{Read/*, Write*/}, 
    panic, 
    path::Path
};
use std::ops::Deref;

use chrono::Utc;
use flaken::Flaken;
use itertools::Itertools;
use rayon::prelude::*;
use serde::Deserialize;
use serenity::{async_trait, framework::standard::{DispatchError, Reason}, model::{channel::Reaction, prelude::{
        Activity, 
        Ready
    }}, utils::Color, prelude::GatewayIntents};
use serenity::client::{
    Client, 
    Context, 
    EventHandler
};
use serenity::model::channel::Message;
use serenity::model::id::UserId;
use serenity::framework::standard::{
    Args,
    StandardFramework,
    CommandResult,
    macros::{group, help, /*check,*/ hook}, 
    HelpOptions, 
    CommandGroup, 
    help_commands,
};
use serenity::prelude::TypeMapKey;
use serenity::utils::MessageBuilder;
use serenity::http::Http;

use pokerust::{Endpoint, PokemonForm, Move};

use curl::easy::{Easy2, Handler as CurlHandler, WriteError};

use regex::Regex;

use serde_json::Value;

use log::*;
use simplelog::*;

mod commands;
use commands::{
    general::*,
    manage::*,
    coach::*,
};

mod data;
use data::*;

mod func;
use func::*;

pub(crate) const EMBED_COLOR: Color = Color::new(0x9c5ac5);

pub(crate) const WATER_EMOJI: &str = "<:Water:797918484721434676>";
pub(crate) const STEEL_EMOJI: &str = "<:Steel:797918484994457600>";
pub(crate) const PSYCHIC_EMOJI: &str = "<:Psychic:797918485028143164>";
pub(crate) const ROCK_EMOJI: &str = "<:Rock:797918485077950474>";
pub(crate) const POISON_EMOJI: &str = "<:Poison:797918484742668329>";
pub(crate) const NORMAL_EMOJI: &str = "<:Normal:797918485065367552>";
pub(crate) const ICE_EMOJI: &str = "<:Ice:797918485010579466>";
pub(crate) const GROUND_EMOJI: &str = "<:Ground:797918484935213076>";
pub(crate) const GRASS_EMOJI: &str = "<:Grass:797918484917911593>";
pub(crate) const GHOST_EMOJI: &str = "<:Ghost:797918484641873962>";
pub(crate) const FLYING_EMOJI: &str = "<:Flying:797918484897595422>";
pub(crate) const FIRE_EMOJI: &str = "<:Fire:797918485178351616>";
pub(crate) const FIGHTING_EMOJI: &str = "<:Fighting:797918484670447617>";
pub(crate) const FAIRY_EMOJI: &str = "<:Fairy:797918484834287637>";
pub(crate) const ELECTRIC_EMOJI: &str = "<:Electric:797918485287534623>";
pub(crate) const DRAGON_EMOJI: &str = "<:Dragon:797918484524826648>";
pub(crate) const DARK_EMOJI: &str = "<:Dark:797918484678967307>";
pub(crate) const BUG_EMOJI: &str = "<:Bug:797918484671627314>";

pub(crate) const TYPE_NAMES: [&str; 19] = [
    "", "Normal", "Fighting", "Flying", "Poison", "Ground", "Rock",
    "Bug", "Ghost", "Steel", "Fire", "Water", "Grass",
    "Electric", "Psychic", "Ice", "Dragon", "Dark", "Fairy"
];

pub(crate) const TYPE_ABBR: [&str; 19] = [
    "", "NRM", "FGH", "FLY", "PSN", "GND", "RCK",
    "BUG", "GHS", "STL", "FIR", "WTR", "GRS",
    "ELE", "PSY", "ICE", "DRG", "DRK", "FAY"
];

pub(crate) fn emoji_from_type_id(type_id: u8) -> &'static str {
    match type_id {
        1 => NORMAL_EMOJI,
        2 => FIGHTING_EMOJI,
        3 => FLYING_EMOJI,
        4 => POISON_EMOJI,
        5 => GROUND_EMOJI,
        6 => ROCK_EMOJI,
        7 => BUG_EMOJI,
        8 => GHOST_EMOJI,
        9 => STEEL_EMOJI,
        10 => FIRE_EMOJI,
        11 => WATER_EMOJI,
        12 => GRASS_EMOJI,
        13 => ELECTRIC_EMOJI,
        14 => PSYCHIC_EMOJI,
        15 => ICE_EMOJI,
        16 => DRAGON_EMOJI,
        17 => DARK_EMOJI,
        18 => FAIRY_EMOJI,
        _ => "bruh"
    }
}

#[group]
#[commands(pokedex, drafts)]
struct General;

#[group]
#[commands(join, status, mon, pick, picks, setmy, bank, team, teams, trade, _match, matches, coaches, playoffs)]
struct Coach;

#[group]
#[prefixes("draftmanage", "draftman", "dman", "dm")]
#[commands(new, modify, default, add, addmon, award, start, playoff, finish, newseason, moderator, submit, modpick, fixpick, setpicks, cache, bonustrade, bonuspick, bonusmega, dropout)]
struct DraftManagement;

#[help]
#[command_not_found_text = "Could not find: `{}`."]
#[max_levenshtein_distance(3)]
#[lacking_permissions = "Hide"]
#[lacking_role = "Nothing"]
#[lacking_conditions = "Nothing"]
#[wrong_channel = "Strike"]
#[embed_success_colour("#9c5ac5")]
#[embed_error_colour("#9c5ac5")]
async fn help(
    context: &Context,
    msg: &Message,
    args: Args,
    help_options: &'static HelpOptions,
    groups: &[&'static CommandGroup],
    owners: HashSet<UserId>
) -> CommandResult {
    let _ = help_commands::with_embeds(context, msg, args, help_options, groups, owners).await;
    Ok(())
}

#[hook]
async fn before(ctx: &Context, msg: &Message, _command_name: &str) -> bool {
    let chill_bypass = if let Ok(c) = msg.channel(ctx).await {
        if let Some(g) = c.guild() {
            let chill_present = g.members(ctx).await.unwrap().iter().filter(|m| *m.user.id.as_u64() == 211047373495992320).exactly_one().map_or(false, |_| true);
            msg.content.starts_with('!') && chill_present
        }
        else {
            false
        }
    }
    else {
        false
    };

    if chill_bypass {
        false
    }
    else {
        let _ = ctx.http.broadcast_typing(msg.channel_id.as_u64().clone()).await;
        
        true
    }
}

#[hook]
async fn unknown_command(ctx: &Context, msg: &Message, unknown_command_name: &str) {
    let chill_bypass = if let Ok(c) = msg.channel(ctx).await {
        if let Some(g) = c.guild() {
            let chill_present = g.members(ctx).await.unwrap().iter().filter(|m| *m.user.id.as_u64() == 211047373495992320).exactly_one().map_or(false, |_| true);
            msg.content.starts_with('!') && chill_present
        }
        else {
            false
        }
    }
    else {
        false
    };

    if !chill_bypass {
        if !unknown_command_name.starts_with('!') && !unknown_command_name.starts_with('~') {
            if !msg.is_private() || {
                let l = msg.content.to_lowercase();
                !l.contains("thank you")
            } {
                let _ = msg.channel_id.say(ctx, MessageBuilder::new()
                .push("Unknown command: ")
                .push_bold_safe(unknown_command_name)
                .build()).await;
            }
        }
    }
}

#[hook]
async fn dispatch_error(ctx: &Context, msg: &Message, error: DispatchError, _command_name: &str) {
    if let DispatchError::CheckFailed(check, reason) = error {
        match reason {
            Reason::User(y) => {
                let _ = msg.reply(ctx, y).await;
            },
            Reason::Log(y) => {
                warn!("{}", y);
            },
            Reason::UserAndLog { user, log } => {
                warn!("{}", log);
                let _ = msg.reply(ctx, user).await;
            },
            Reason::Unknown | _ => {
                warn!("Check {} failed for unknown reasons", check);
            },
        }
    }
    else {
        match error {
            DispatchError::OnlyForGuilds => {
                let _ = msg.reply(ctx, "That command must be run in a server!").await;
            },
            _ => {
                warn!("{:?}", error);
            }
        }
    }
}

struct SmogonTierList;

impl TypeMapKey for SmogonTierList {
    type Value = HashMap<i16, String>;
}

struct DoublesTierList;

impl TypeMapKey for DoublesTierList {
    type Value = HashMap<i16, String>;
}

struct PPATierList;

impl TypeMapKey for PPATierList {
    type Value = HashMap<i16, (String, Option<u8>)>;
}

struct Learnsets;

impl TypeMapKey for Learnsets {
    type Value = HashMap<u8, HashMap<i16, Vec<i16>>>;
}

struct DraftRegistry;

impl TypeMapKey for DraftRegistry {
    type Value = HashMap<String, Draft>;
}

struct DefaultDrafts;

impl TypeMapKey for DefaultDrafts {
    type Value = HashMap<u64, String>;
}

struct Handler;

#[async_trait]
impl EventHandler for Handler {
    async fn message(&self, ctx: Context, msg: Message) {
        let ctx = &ctx;
        let msg = &msg;
        let sid = msg.guild_id.map_or(0, |g| g.as_u64().to_owned());
        let cid = msg.channel_id.as_u64().to_owned();
        let (isc, draft_id) = ctx.is_porygon_channel(&sid, &cid).await;
        if isc && *msg.author.id.as_u64() == 692091256477581423 {
            match ctx.create_match(draft_id.clone(), msg.content.clone()).await {
                Ok(mut m) => {
                    match ctx.commit_match(&draft_id, m.clone(), false).await {
                        Ok(p) => {
                            match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                                Ok(a_c) => {
                                    if p.is_some() {
                                        m.playoffs = true;
                                    }
                                    let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
                                    let e = ctx.match_embed(&draft_id, &m, s_id.clone()).await;
                                    let _ = a_c.id().send_message(ctx, |m| {
                                        m.set_embed(e);
                                        m
                                    }).await;
                                    if p.is_some() {
                                        let p = p.as_ref().unwrap();
                                        let e = ctx.playoff_match_embed(p, &draft_id, s_id).await;
                                        let _ = a_c.id().send_message(ctx, |m| {
                                            m.set_embed(e);
                                            m
                                        }).await;
                                    }
                                }
                                Err(y) => {
                                    error!("Porygon: Failed getting draft announcement channel: {}", y);
                                }
                            }
                        },
                        Err(y) => {
                            match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                                Ok(a_c) => {
                                    let _ = a_c.id().say(ctx, y).await;
                                }
                                Err(y) => {
                                    error!("Porygon: Failed getting draft announcement channel: {}", y);
                                }
                            }
                        }
                    }
                },
                Err(y) => {
                    match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                        Ok(a_c) => {
                            let _ = a_c.id().say(ctx, y).await;
                        }
                        Err(y) => {
                            error!("Porygon: Failed getting draft announcement channel: {}", y);
                        }
                    }
                }
            }
            /*info!("Processing Porygon Match Report");
            let date = Utc::now();
            let mut content = msg.content.clone();
            content = Regex::new(r"\n+").unwrap().replace_all(&content, "\n").into_owned();
            //debug!("{}", content);
            let winner_scores = Regex::new(r"Result:.*\|\| ?(?P<winner>.+) won (?P<winnerscore>[0-9])\-(?P<loserscore>[0-9])").unwrap();
            let players_mons = Regex::new(r"\*\*(?P<player1>.+)\*\*:.*\n(?P<p1mons>(.+,[0-9]+,[0-9]+,[0-9]+\n){1,6})(\|\|)? \n(\|\|)?\*\*(?P<player2>.+)\*\*:.*\n(?P<p2mons>(.+,[0-9]+,[0-9]+,[0-9]+\n){1,6})").unwrap();
            let replay = Regex::new(r"Replay.*(?P<replay>http.+)>\n").unwrap();

            let replay_cap = replay.captures(content.as_str()).unwrap();
            let win_scores_cap = winner_scores.captures(content.as_str()).unwrap();
            let p_mons_cap = players_mons.captures(content.as_str()).unwrap();
        
            let winner_string = win_scores_cap.name("winner").unwrap().as_str().to_lowercase();
            let winner = winner_string.as_str();
            let winscore = win_scores_cap.name("winnerscore").unwrap().as_str();
            let losescore = win_scores_cap.name("loserscore").unwrap().as_str();
        
            let p1user = p_mons_cap.name("player1").unwrap().as_str().to_lowercase();
            let p1mons = p_mons_cap.name("p1mons").unwrap().as_str();
            let p2user = p_mons_cap.name("player2").unwrap().as_str().to_lowercase();
            let p2mons = p_mons_cap.name("p2mons").unwrap().as_str();
        
            let replay = replay_cap.name("replay").unwrap().as_str().to_owned();
            
            let (p1e, coach1) = ctx.check_draft_showdown(&draft_id, &p1user).await;
            let (p2e, coach2) = ctx.check_draft_showdown(&draft_id, &p2user).await;
            if p1e && p2e {
                info!("Porygon: Both Showdown names match, proceeding");
                let p1mons = p1mons.split('\n');
                let winscore = winscore.parse::<u8>().unwrap();
                debug!("Porygon: Player 1 team size: {}; Score: {}", p1mons.clone().count(), winscore);
                let deaths1 = (p1mons.clone().count() as u8 - 1) - winscore;
                let mut team1: Vec<MatchPokemon> = Vec::new();
                for m in p1mons {
                    if m != "" {
                        let csv: Vec<String> = m.split(',').map(|s| s.to_owned()).collect();
                        let mon = {
                            let mut m = csv[0].to_lowercase();
                            m = m.replace(" ", "-");
                            m = Regex::new(r"[\p{Punct}--[\-]]+").unwrap().replace_all(m.as_str(), "").into_owned();
                            m = Regex::new(r"-$").unwrap().replace_all(m.as_str(), "").into_owned();
    
                            m
                        };
                        debug!("Porygon: Attempting to include Pokemon '{}' in match result", mon);
                        let ak = csv[1].parse::<u8>().unwrap();
                        let pk = csv[2].parse::<u8>().unwrap();
                        let d = csv[3].parse::<u8>().unwrap();
                        match PokemonForm::from_name_exhaustive(mon.as_str()) {
                            Ok(f) => {
                                if ctx.check_coach_has_mon(draft_id.clone(), &coach1, &f.id).await {
                                    let p = MatchPokemon {
                                        id: f.id,
                                        active_kills: ak,
                                        passive_kills: pk,
                                        deaths: d,
                                    };
                                    team1.push(p);
                                }
                                else {
                                    let (o_id, m_id) = ctx.get_draft_owner_and_moderators(&draft_id).await.unwrap();
                                    let owner = ctx.http.get_user(o_id).await;
                                    match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                                        Ok(a_c) => {
                                            let mut mb = MessageBuilder::new();
                                            if owner.is_ok() { mb.mention(&owner.unwrap()); }
                                            if m_id.is_some() {
                                                mb.push(" ");
                                                let m_id = m_id.unwrap();
                                                let guild = msg.guild(ctx).await.unwrap();
                                                let m_role = guild.roles.iter()
                                                    .filter(|(r, _)| r.as_u64() == &m_id).map(|(r, _)| r).next().unwrap();
                                                mb.mention(m_role);
                                            }
                                            mb.push(": Porygon attempted to submit a match with an incorrect Pokémon on ");
                                            mb.push_safe(p1user);
                                            mb.push_line("'s team. This could just be an oversight on either Genesect or Porygon's part, or that coach brought a Pokémon they didn't draft.");
                                            mb.push("Problematic Pokémon: ");
                                            mb.push_mono_line(mon);
                                            mb.push_line("The full Porygon output will be repeated here in the next message; please fix it and then re-submit the match, or tell your coaches to rematch.");

                                            let _ = a_c.id().say(ctx, mb.build()).await;
                                            let _ = a_c.id().say(ctx, MessageBuilder::new()
                                                .push_bold_line("Problematic Porygon Output:")
                                                .push_codeblock(content, None)
                                                .build()).await;
                                        }
                                        Err(y) => {
                                            error!("Porygon: Failed getting draft announcement channel: {}", y);
                                        }
                                    }
                                    return;
                                }
                            },
                            Err(_) => {
                                let (o_id, m_id) = ctx.get_draft_owner_and_moderators(&draft_id).await.unwrap();
                                let owner = ctx.http.get_user(o_id).await;
                                match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                                    Ok(a_c) => {
                                        let mut mb = MessageBuilder::new();
                                        if owner.is_ok() { mb.mention(&owner.unwrap()); }
                                        if m_id.is_some() {
                                            mb.push(" ");
                                            let m_id = m_id.unwrap();
                                            let guild = msg.guild(ctx).await.unwrap();
                                            let m_role = guild.roles.iter()
                                                .filter(|(r, _)| r.as_u64() == &m_id).map(|(r, _)| r).next().unwrap();
                                            mb.mention(m_role);
                                        }
                                        mb.push_line(": Porygon attempted to submit a match with an incorrect or malformed Pokémon name. (Or my author has forgotten an edge case. oops)");
                                        mb.push("Problematic name: ");
                                        mb.push_mono_line(mon);
                                        mb.push_line("The full Porygon output will be repeated here in the next message; please fix it and then re-submit the match.");

                                        let _ = a_c.id().say(ctx, mb.build()).await;
                                        let _ = a_c.id().say(ctx, MessageBuilder::new()
                                            .push_bold_line("Problematic Porygon Output:")
                                            .push_codeblock(content, None)
                                            .build()).await;
                                    }
                                    Err(y) => {
                                        error!("Porygon: Failed getting draft announcement channel: {}", y);
                                    }
                                }
                                return;
                            },
                        }
                    }
                }
                let p2mons = p2mons.split('\n');
                let losescore = losescore.parse::<u8>().unwrap();
                debug!("Porygon: Player 2 team size: {}; Score: {}", p2mons.clone().count(), losescore);
                let deaths2 = (p2mons.clone().count() as u8 - 1) - losescore;
                let mut team2: Vec<MatchPokemon> = Vec::new();
                for m in p2mons {
                    if m != "" {
                        let csv: Vec<String> = m.split(',').map(|s| s.to_owned()).collect();
                        let mon = {
                            let mut m = csv[0].to_lowercase();
                            m = m.replace(" ", "-");
                            m = Regex::new(r"[\p{Punct}--[\-]]+").unwrap().replace_all(m.as_str(), "").into_owned();
                            m = Regex::new(r"-$").unwrap().replace_all(m.as_str(), "").into_owned();
    
                            m
                        };
                        debug!("Porygon: Attempting to include Pokemon '{}' in match result", mon);
                        let ak = csv[1].parse::<u8>().unwrap();
                        let pk = csv[2].parse::<u8>().unwrap();
                        let d = csv[3].parse::<u8>().unwrap();
                        match PokemonForm::from_name_exhaustive(mon.as_str()) {
                            Ok(f) => {
                                if ctx.check_coach_has_mon(draft_id.clone(), &coach2, &f.id).await {
                                    let p = MatchPokemon {
                                        id: f.id,
                                        active_kills: ak,
                                        passive_kills: pk,
                                        deaths: d,
                                    };
                                    team2.push(p);
                                }

                                else {
                                    let (o_id, m_id) = ctx.get_draft_owner_and_moderators(&draft_id).await.unwrap();
                                    let owner = ctx.http.get_user(o_id).await;
                                    match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                                        Ok(a_c) => {
                                            let mut mb = MessageBuilder::new();
                                            if owner.is_ok() { mb.mention(&owner.unwrap()); }
                                            if m_id.is_some() {
                                                mb.push(" ");
                                                let m_id = m_id.unwrap();
                                                let guild = msg.guild(ctx).await.unwrap();
                                                let m_role = guild.roles.iter()
                                                    .filter(|(r, _)| r.as_u64() == &m_id).map(|(r, _)| r).next().unwrap();
                                                mb.mention(m_role);
                                            }
                                            mb.push(": Porygon attempted to submit a match with an incorrect Pokémon on ");
                                            mb.push_safe(p2user);
                                            mb.push_line("'s team. This could just be an oversight on either Genesect or Porygon's part, or that coach brought a Pokémon they didn't draft.");
                                            mb.push("Problematic Pokémon: ");
                                            mb.push_mono_line(mon);
                                            mb.push_line("The full Porygon output will be repeated here in the next message; please fix it and then re-submit the match, or tell your coaches to rematch.");

                                            let _ = a_c.id().say(ctx, mb.build()).await;
                                            let _ = a_c.id().say(ctx, MessageBuilder::new()
                                                .push_bold_line("Problematic Porygon Output:")
                                                .push_codeblock(content, None)
                                                .build()).await;
                                        }
                                        Err(y) => {
                                            error!("Porygon: Failed getting draft announcement channel: {}", y);
                                        }
                                    }
                                    return;
                                }
                            },
                            Err(_) => {
                                let (o_id, m_id) = ctx.get_draft_owner_and_moderators(&draft_id).await.unwrap();
                                let owner = ctx.http.get_user(o_id).await;
                                match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                                    Ok(a_c) => {
                                        let mut mb = MessageBuilder::new();
                                        if owner.is_ok() { mb.mention(&owner.unwrap()); }
                                        if m_id.is_some() {
                                            mb.push(" ");
                                            let m_id = m_id.unwrap();
                                            let guild = msg.guild(ctx).await.unwrap();
                                            let m_role = guild.roles.iter()
                                                .filter(|(r, _)| r.as_u64() == &m_id).map(|(r, _)| r).next().unwrap();
                                            mb.mention(m_role);
                                        }
                                        mb.push_line(": Porygon attempted to submit a match with an incorrect or malformed Pokémon name. (Or my author has forgotten an edge case. oops)");
                                        mb.push("Problematic name: ");
                                        mb.push_mono_line(mon);
                                        mb.push_line("The full Porygon output will be repeated here in the next message; please fix it and then re-submit the match.");

                                        let _ = a_c.id().say(ctx, mb.build()).await;
                                        let _ = a_c.id().say(ctx, MessageBuilder::new()
                                            .push_bold_line("Problematic Porygon Output:")
                                            .push_codeblock(content, None)
                                            .build()).await;
                                    }
                                    Err(y) => {
                                        error!("Porygon: Failed getting draft announcement channel: {}", y);
                                    }
                                }
                                return;
                            },
                        }
                    }
                }
                let winner_is_1 = winner.to_owned() == p1user.clone();
                info!("Porygon: Match parse success. Winner: `{}`; Player 1: `{}`; Player 1 won: {}", winner, p1user, winner_is_1);
                let id = ctx.get_next_match_id(&draft_id).await as u16;
                let mut flake = Flaken::default().node(1).epoch(1613940319944);
                let uuid = flake.next();

                let m = Match {
                    uuid,
                    id,
                    coach1,
                    coach2,
                    team1,
                    team2,
                    winner_is_1,
                    c1_score: winscore,
                    c2_score: losescore,
                    deaths1,
                    deaths2,
                    replay,
                    date
                };

                ctx.commit_match(&draft_id, m.clone()).await;
                
                match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                    Ok(a_c) => {
                        let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
                        let e = ctx.match_embed(&draft_id, &m, s_id).await;
                        let _ = a_c.id().send_message(ctx, |m| {
                            m.set_embed(e);
                            m
                        }).await;
                    }
                    Err(y) => {
                        error!("Porygon: Failed getting draft announcement channel: {}", y);
                    }
                }
            }
            else {
                let (o_id, m_id) = ctx.get_draft_owner_and_moderators(&draft_id).await.unwrap();
                let owner = ctx.http.get_user(o_id).await;
                match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                    Ok(a_c) => {
                        let mut mb = MessageBuilder::new();
                        if owner.is_ok() { mb.mention(&owner.unwrap()); }
                        if m_id.is_some() {
                            mb.push(" ");
                            let m_id = m_id.unwrap();
                            let guild = msg.guild(ctx).await.unwrap();
                            let m_role = guild.roles.iter()
                                .filter(|(r, _)| r.as_u64() == &m_id).map(|(r, _)| r).next().unwrap();
                            mb.mention(m_role);
                        }
                        mb.push_line(": Porygon attempted to submit a match with Showdown accounts that aren't registered in this draft.");
                        mb.push_line("The full Porygon output will be repeated here in the next message; please fix it and then re-submit the match.");
                        mb.push_bold_line("In the future, ensure your players are only using one Showdown account for draft matches.");

                        let _ = a_c.id().say(ctx, mb.build()).await;
                        let _ = a_c.id().say(ctx, MessageBuilder::new()
                            .push_bold_line("Problematic Porygon Output:")
                            .push_codeblock(content, None)
                            .build()).await;
                    }
                    Err(y) => {
                        error!("Porygon: Failed getting draft announcement channel: {}", y);
                    }
                }
            }
            */
        }

        if msg.is_private() && {
            let l = msg.content.to_lowercase();
            l.contains("thank you")
        } {
            let _ = msg.channel_id.say(ctx, "<:genesect:821101737346990110><a:bubble_dots:830918993215881260> `You're welcome.`").await;
        }
    }

    async fn ready(&self, ctx: Context, ready: Ready) {
        info!("{} is connected!", ready.user.name);
        ctx.shard.set_activity(Some(Activity::listening("g~help")));
    }

    async fn reaction_add(&self, ctx: Context, reax: Reaction) {
        let reac = reax.emoji.to_string();
        debug!("reac it is {}", &reac);

        if reac.contains('\u{1f4cc}') {
            if let Ok(r_u) = reax.user(&ctx).await {
                if let Ok(m) = reax.message(&ctx).await {
                    if let Ok(s) = ctx.http.get_current_user().await {
                        if m.author.id == s.id {
                            if let Ok(c) = reax.channel(&ctx).await {
                                if let Some(gc) = c.guild() {
                                    if let Ok(r_m) = ctx.http.get_member(gc.guild_id.0, r_u.id.0).await {
                                        if let Ok(p) = r_m.permissions(&ctx) {
                                            if let Ok(app) = ctx.http.get_current_application_info().await {
                                                let owner = app.owner.id;

                                                if p.administrator() || r_u.id == owner {
                                                    debug!("pinin");
                                                    match m.pin(&ctx).await {
                                                        Ok(_) => {
                                                            debug!("pinned");
                                                        },
                                                        Err(y) => {
                                                            error!("couldn't pin: {}", y);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}

#[tokio::main]
async fn main() {
    dotenv::dotenv().expect("Failed to load .env file");

    CombinedLogger::init(
        vec![
            TermLogger::new(LevelFilter::Info, ConfigBuilder::new()
                .add_filter_allow_str("genesect")
                .build(),
                TerminalMode::Mixed,
                ColorChoice::Auto),
            WriteLogger::new(LevelFilter::Debug, ConfigBuilder::new()
                .add_filter_allow_str("genesect")
                .build(),
                OpenOptions::new().append(true).create(true).open(format!("genesect_{}.log", Utc::now().format("%Y-%m-%d"))).unwrap()),
            WriteLogger::new(LevelFilter::Debug, ConfigBuilder::new()
                .build(),
                OpenOptions::new().append(true).create(true).open(format!("genesect_debug_{}.log", Utc::now().format("%Y-%m-%d"))).unwrap()),
        ]
    ).unwrap();

    panic::set_hook(Box::new(|panic_info| {
        let (filename, line) =
            panic_info.location().map(|loc| (loc.file(), loc.line()))
                .unwrap_or(("<unknown>", 0));
    
        let cause = panic_info.payload().downcast_ref::<String>().map(String::deref);
    
        let cause = cause.unwrap_or_else(||
            panic_info.payload().downcast_ref::<&str>().map(|s| *s)
                .unwrap_or("<cause unknown>")
        );
    
        error!("A panic occurred at {}:{}: {}", filename, line, cause);
    }));

    info!("Genesect starting");

    let token = env::var("DISCORD_TOKEN").expect(
        "Expected a token in the environment",
    );

    let http = Http::new(&token);

    let (owners, bot_id) = match (http.get_current_application_info().await, http.get_current_user().await) {
        (Ok(app), Ok(user)) => {
            let mut owners = HashSet::new();
            owners.insert(app.owner.id);

            (owners, user.id)
        },
        (Err(why), Err(_)) | (Err(why), _) | (_, Err(why)) => panic!("Could not access application info: {:?}", why),
    };

    let framework = StandardFramework::new()
        .configure(|c| c.prefixes(vec!["!", "g~", "g!"])
        .case_insensitivity(true)
        .on_mention(Some(bot_id))
        .no_dm_prefix(true)
        .owners(owners))
        .before(before)
        .unrecognised_command(unknown_command)
        .on_dispatch_error(dispatch_error)
        .help(&HELP)
        .group(&GENERAL_GROUP)
        .group(&COACH_GROUP)
        .group(&DRAFTMANAGEMENT_GROUP);

    // Login with a bot token from the environment
    let token = env::var("DISCORD_TOKEN").expect("token");
    let mut client = Client::builder(token, GatewayIntents::all())
        .event_handler(Handler)
        .framework(framework)
        .await
        .expect("Error creating client");

    struct Collector(Vec<u8>);

    impl CurlHandler for Collector {
        fn write(&mut self, data: &[u8]) -> Result<usize, WriteError> {
            self.0.extend_from_slice(data);
            Ok(data.len())
        }
    }
    
    let _gen8_tiers_json: Value = {
        let mut easy = Easy2::new(Collector(Vec::new()));
        easy.get(true).unwrap();
        easy.url("https://raw.githubusercontent.com/smogon/pokemon-showdown/master/data/mods/gen8/formats-data.ts").unwrap();
        easy.perform().unwrap();
        
        let contents = easy.get_ref();
        let mut data = String::from_utf8_lossy(&contents.0).into_owned();
        data = data.replace("export const FormatsData: {[k: string]: SpeciesFormatsData} = {", "{");
        data = data.replace("};", "}");
        data = data.replace("(", "");
        data = data.replace(")", "");
        data = Regex::new(r"\{\n\s+").unwrap().replace_all(&data, "{\n\"").into_owned();
        data = Regex::new(r":\s+").unwrap().replace_all(&data, "\":").into_owned();
        data = Regex::new(r"\},\n\s+").unwrap().replace_all(&data, "},\n\"").into_owned();
        data = Regex::new(r",\n\s+").unwrap().replace_all(&data, ",\n\"").into_owned();
        data = Regex::new(r#""\},"#).unwrap().replace_all(&data, "},").into_owned();
        data = Regex::new(r",\n\}").unwrap().replace_all(&data, "\n}").into_owned();
        data = Regex::new(r"//.*\n\s+").unwrap().replace_all(&data, "\n\"").into_owned();

        serde_json::from_str(&data).unwrap()
    };
    
    let bdsp_json: Value = {
        let mut easy = Easy2::new(Collector(Vec::new()));
        easy.get(true).unwrap();
        easy.url("https://raw.githubusercontent.com/smogon/pokemon-showdown/master/data/mods/gen8bdsp/learnsets.ts").unwrap();
        easy.perform().unwrap();
        
        let contents = easy.get_ref();
        let mut data = String::from_utf8_lossy(&contents.0).into_owned();
        data = data.replace("/* eslint-disable max-len */\n\nexport const Learnsets: {[k: string]: ModdedLearnsetData} = {", "{");
        data = data.replace("};", "}");
        data = Regex::new(r"\n\s+eventData.*\n\s+\{.*\n\s+.*\n\s+.*").unwrap().replace_all(&data, "").into_owned();
        data = Regex::new(r"\{\n\s+").unwrap().replace_all(&data, "{\n\"").into_owned();
        data = Regex::new(r":\s+").unwrap().replace_all(&data, "\":").into_owned();
        data = Regex::new(r"\},\n\s+").unwrap().replace_all(&data, "},\n\"").into_owned();
        data = Regex::new(r",\n\s+").unwrap().replace_all(&data, ",\n\"").into_owned();
        data = Regex::new(r#""\},"#).unwrap().replace_all(&data, "},").into_owned();
        data = Regex::new(r",\n\}").unwrap().replace_all(&data, "\n}").into_owned();
        data = Regex::new(r#"\{\n"learnset":\{\n"#).unwrap().replace_all(&data, "[").into_owned();
        data = Regex::new(r"\n\}\n\},").unwrap().replace_all(&data, "],").into_owned();
        data = Regex::new(r":\[[^:\n\]]*\]").unwrap().replace_all(&data, "").into_owned();
        data = Regex::new(r"\n\}\n\}").unwrap().replace_all(&data, "\n]").into_owned();
        data = Regex::new(r"\n\},").unwrap().replace_all(&data, "],").into_owned();
        data = Regex::new(r#"".*":\{\n.*\n"#).unwrap().replace_all(&data, "").into_owned();
        data = Regex::new(r".*:\{\],\n").unwrap().replace_all(&data, "").into_owned();
        
        /*let mut file = match File::create(Path::new("test.json")) {
            Err(why) => unreachable!(),
            Ok(file) => file,
        };

        match file.write_all(data.as_bytes()) {
            Err(why) => unreachable!(),
            Ok(_) => debug!("successfully wrote to test.json"),
        }*/

        serde_json::from_str(&data).unwrap()
    };
    
    let gen9_tiers_json: Value = {
        let mut easy = Easy2::new(Collector(Vec::new()));
        easy.get(true).unwrap();
        easy.url("https://raw.githubusercontent.com/smogon/pokemon-showdown/master/data/formats-data.ts").unwrap();
        easy.perform().unwrap();
        
        let contents = easy.get_ref();
        let mut data = String::from_utf8_lossy(&contents.0).into_owned();
        data = data.replace("export const FormatsData: {[k: string]: SpeciesFormatsData} = {", "{");
        data = data.replace("};", "}");
        data = data.replace("(", "");
        data = data.replace(")", "");
        data = Regex::new(r"\{\n\s+").unwrap().replace_all(&data, "{\n\"").into_owned();
        data = Regex::new(r":\s+").unwrap().replace_all(&data, "\":").into_owned();
        data = Regex::new(r"\},\n\s+").unwrap().replace_all(&data, "},\n\"").into_owned();
        data = Regex::new(r",\n\s+").unwrap().replace_all(&data, ",\n\"").into_owned();
        data = Regex::new(r#""\},"#).unwrap().replace_all(&data, "},").into_owned();
        data = Regex::new(r",\n\}").unwrap().replace_all(&data, "\n}").into_owned();
        data = Regex::new(r"//.*\n\s+").unwrap().replace_all(&data, "\n\"").into_owned();

        serde_json::from_str(&data).unwrap()
    };

    let forms = PokemonForm::list(0, 2000).unwrap().results;
    let moves = Move::list(0, 1500).unwrap().results;
    let mut single_tiers: HashMap<i16, String> = HashMap::new();
    let mut double_tiers: HashMap<i16, String> = HashMap::new();
    let mut bdsp_learnsets: HashMap<i16, Vec<i16>> = HashMap::new();
    for f in &forms {
        if !f.name.ends_with("-gmax") {
            let p: Vec<&str> = f.name.split("-").collect();
            let poke = p.first().unwrap().to_string();
            let form = if f.name.ends_with("breed") {
                let form = f.name.replace("-", "");
                let regex = Regex::new(r"(?P<realform>.+)breed").unwrap();
                let cap = regex.captures(form.as_str()).unwrap();
                let out = cap.name("realform").unwrap().as_str().to_lowercase();
                out
            } else { f.name.replace("-", "") };
            let u: Vec<&str> = f.url.split("/").collect();
            let id = u[6].parse::<i16>().unwrap();
            let f_data = gen9_tiers_json[&form].clone();
            let p_data = gen9_tiers_json[&poke].clone();
            debug!("tiering poke {} form {}, fdata {:?} pdata {:?}", poke, form, &f_data, &p_data);
            if f_data != Value::Null {
                let s_tier = f_data["tier"].clone();
                let d_tier = f_data["doublesTier"].clone();
                if s_tier != Value::Null || d_tier != Value::Null {
                    if s_tier != Value::Null {
                        debug!("{}: S{}", form, s_tier.to_string());
                        single_tiers.insert(id, s_tier.as_str().unwrap().to_owned());
                    }
                    if d_tier != Value::Null {
                        debug!("{}: D{}", form, d_tier.to_string());
                        double_tiers.insert(id, d_tier.as_str().unwrap().to_owned().replace("D", ""));
                    }
                }
                let b_data = bdsp_json[&form].clone();
                let mut learnset = Vec::new();
                if b_data != Value::Null {
                    if let Some(learnset_names) = b_data.as_array() {
                        let l = learnset_names.iter().map(|v| v.as_str().unwrap().to_owned()).collect::<Vec<String>>();
                        for m in &moves {
                            let dedashed = m.name.replace("-", "");
                            let u: Vec<&str> = m.url.split("/").collect();
                            let id = u[6].parse::<i16>().unwrap();
                            if l.contains(&dedashed) {
                                learnset.push(id);
                            }
                        }
                    }
                }
                if f.name.contains("rotom-") {
                    let b_data = bdsp_json["rotom"].clone();
                    if b_data != Value::Null {
                        if let Some(learnset_names) = b_data.as_array() {
                            let l = learnset_names.iter().map(|v| v.as_str().unwrap().to_owned()).collect::<Vec<String>>();
                            for m in &moves {
                                let dedashed = m.name.replace("-", "");
                                let u: Vec<&str> = m.url.split("/").collect();
                                let id = u[6].parse::<i16>().unwrap();
                                if l.contains(&dedashed) {
                                    learnset.push(id);
                                }
                            }
                        }
                    }
                }
                bdsp_learnsets.insert(id, learnset);
            }
            else if p_data != Value::Null {
                let s_tier = p_data["tier"].clone();
                let d_tier = p_data["doublesTier"].clone();
                if s_tier != Value::Null || d_tier != Value::Null {
                    if s_tier != Value::Null {
                        debug!("{}: S{}", form, s_tier.to_string());
                        single_tiers.insert(id, s_tier.as_str().unwrap().to_owned());
                    }
                    if d_tier != Value::Null {
                        debug!("{}: D{}", form, d_tier.to_string());
                        double_tiers.insert(id, d_tier.as_str().unwrap().to_owned().replace("D", ""));
                    }
                }
                let b_data = bdsp_json[&poke].clone();
                let mut learnset = Vec::new();
                if b_data != Value::Null {
                    if let Some(learnset_names) = b_data.as_array() {
                        let l = learnset_names.iter().map(|v| v.as_str().unwrap().to_owned()).collect::<Vec<String>>();
                        for m in &moves {
                            let dedashed = m.name.replace("-", "");
                            let u: Vec<&str> = m.url.split("/").collect();
                            let id = u[6].parse::<i16>().unwrap();
                            if l.contains(&dedashed) {
                                learnset.push(id);
                            }
                        }
                    }
                }
                bdsp_learnsets.insert(id, learnset);
            }
        }
    }

    /*let testron = ron::ser::to_string_pretty(&bdsp_learnsets, ron::ser::PrettyConfig::new()).unwrap();
    let mut file = match File::create(Path::new("test.ron")) {
        Err(why) => unreachable!(),
        Ok(file) => file,
    };

    match file.write_all(testron.as_bytes()) {
        Err(why) => unreachable!(),
        Ok(_) => debug!("successfully wrote to test.ron"),
    }*/

    let mut learnsets = HashMap::new();
    learnsets.insert(21, bdsp_learnsets);

    #[derive(Debug, Deserialize)]
    struct PPAMon {
        pokemon: String,
        tier: String,
        banned_ability: Option<u8>,
    }

    let mut ppa_tiers: HashMap<i16, (String, Option<u8>)> = HashMap::new();
    let mut rdr = match csv::Reader::from_path("data/ppatiers.csv") {
        Err(why) => panic!("couldn't open ppatiers: {}", why),
        Ok(r) => r,
    };
    for result in rdr.deserialize() {
        let record: PPAMon = result.unwrap();
        let mut matching = forms.iter().filter(|a| a.name == record.pokemon.to_lowercase());
        if matching.clone().count() == 1 {
            let f = matching.next().unwrap();
            let u: Vec<&str> = f.url.split("/").collect();
            let id = u[6].parse::<i16>().unwrap();
            debug!("{}, {}, {:?}", id, &record.tier, &record.banned_ability);
            ppa_tiers.insert(id, (record.tier, record.banned_ability));
        }
    }
    
    match fs::create_dir("drafts/") {
        Ok(_) => info!("Created new drafts dir"),
        Err(_) => debug!("drafts dir already exists")
    }

    let default_drafts = {
        let path = Path::new("data/default_drafts.ron");
        match File::open(path) {
            Err(_) => {
                HashMap::new()
            },
            Ok(mut f) => {
                let mut ron = String::new();
                match f.read_to_string(&mut ron) {
                    Err(_) => HashMap::new(),
                    Ok(_) => {
                        let r: HashMap<u64, String> = ron::from_str(&ron).unwrap_or_default();
                        r
                    },
                }
            }
        }
    };

    {
        let mut data = client.data.write().await;
        data.insert::<SmogonTierList>(single_tiers);
        data.insert::<DoublesTierList>(double_tiers);
        data.insert::<PPATierList>(ppa_tiers);
        data.insert::<Learnsets>(learnsets);
        data.insert::<DefaultDrafts>(default_drafts);
        data.insert::<DraftRegistry>(HashMap::new());

        for entry in fs::read_dir(Path::new("drafts/")).expect("drafts directory doesn't exist") {
            let entry = entry.expect("IO error?");
            let path = entry.path();
            if !path.is_dir() {
                let display = path.display();

                let mut file = match File::open(&path) {
                    Err(why) => panic!("couldn't open {}: {}", display, why),
                    Ok(file) => file,
                };

                let mut ron = String::new();
                match file.read_to_string(&mut ron) {
                    Err(why) => panic!("couldn't read {}: {}", display, why),
                    Ok(_) => debug!("{} contains ron", display),
                }

                let correct = match env::var("GENESECT_UPDATE_IDS") {
                    Ok(_) => true,
                    Err(_) => false,
                };
                let mut d = if correct {
                    let mut d: Draft = ron::from_str(&ron).unwrap();
                    let new_mons: HashMap<i16, DraftPokemon> = d.mons.par_iter().map(|(i, m)| match m.correct() {
                        Some((j, n)) => (j, n),
                        None => (*i, m.clone())
                    }).collect();
                    d.coaches.par_iter_mut().for_each(|(i, c)| {
                        c.mons = new_mons.iter().filter(|(_, m)| m.coach.is_some() && m.coach.as_ref().unwrap() == i).map(|(i, _)| *i).collect();
                    });
                    d.mons = new_mons;
                    if d.past_seasons.is_some() {
                        d.past_seasons.as_mut().unwrap().iter_mut().for_each(|(_, s)| {
                            let new_mons: HashMap<i16, DraftPokemon> = s.mons.par_iter().map(|(i, m)| match m.correct() {
                                Some((j, n)) => (j, n),
                                None => (*i, m.clone())
                            }).collect();
                            s.coaches.par_iter_mut().for_each(|(i, c)| {
                                c.mons = new_mons.iter().filter(|(_, m)| m.coach.is_some() && m.coach.as_ref().unwrap() == i).map(|(i, _)| *i).collect();
                            });
                            s.mons = new_mons;
                        });
                    }
                    d
                }
                else {
                    match ron::from_str::<Draft>(&ron) {
                        Ok(d) => {
                            d
                        },
                        Err(why) => {
                            error!("Can't load {}. Details: {:?}", display, why);
                            continue;
                        }
                    }
                };
                let mut flake = Flaken::default().node(1).epoch(1613940319944);
                for pm in &mut d.playoff_matches {
                    if &pm.challonge_id == &0 {
                        pm.challonge_id = flake.next();
                    }
                }
                d.save();
                /*d.mons = HashMap::new();
                for (i, m) in mons {
                    match m.correct() {
                        Some((j, n)) => d.mons.insert(j, n),
                        None => d.mons.insert(i, m)
                    };
                }*/
                {
                    let registry = data.get_mut::<DraftRegistry>().expect("Expected serverspawns in TypeMap.");
                    registry.insert(d.draft_id.clone(), d);
                }
            }
        }
    }

    // start listening for events by starting a single shard
    if let Err(why) = client.start().await {
        error!("An error occurred while running the client: {:?}", why);
    }
}
